// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file MacroElement.h */

#ifndef AMDIS_MACROELEMENT_H
#define AMDIS_MACROELEMENT_H

#include <deque>
#include <stdio.h>
#include "Boundary.h"
#include "Global.h"
#include "MemoryManager.h"
#include "Projection.h"
#include "FixVec.h"
#include "Serializable.h"

namespace AMDiS {

  class Element;
  class ElInfo;
  class Mesh;

  template<typename T> class WorldVector;
  template<typename T> class DimVec;

  // ============================================================================
  // ===== class MacroElement ===================================================
  // ============================================================================

  /** \ingroup Triangulation
   * \brief
   * MacroElements form the macro triangulation of a Mesh. In a MacroElement
   * geometrical information are stored, which are used in mesh traversal,
   * to calculate the desired information and fill it in an ElInfo object. 
   */ 
  class MacroElement : public Serializable
  {
  public:
    MEMORY_MANAGED(MacroElement);

    /** \brief
     * Creates a new MacroElement. The mesh is needed only to get the dimension
     */
    MacroElement(int dim); 

    /** \brief
     * Destructor.
     */
    virtual ~MacroElement();

    // ===== getting methods ======================================================

    /** \name getting methods
     * \{
     */

    /** \brief
     * Returns \ref index.
     */
    inline int getIndex() const {
      return index; 
    };

    /** \brief
     * Returns ref projection[i]. 
     */
    inline Projection *getProjection(int i) const {
      return projection[i];
    };

    /** \brief
     * Returns \ref el
     */
    inline Element* getElement() const {
      return element; 
    };

    /** \brief
     * Returns the i-th neighbour of this MacroElement \ref neighbour[i]
     */
    inline MacroElement* getNeighbour(int i) const {
      return neighbour[i];
    };

    /** \brief
     * Returns the i-th opp-vertex of this MacroElement \ref oppVertex[i]
     */
    inline char getOppVertex(int i) const {
      return oppVertex[i];
    };

    /** \brief
     * Returns \ref coord[i]
     */
    inline WorldVector<double>& getCoord(int i) {
      return coord[i];    
    };

    /** \brief
     * Returns \ref coord
     */
    inline FixVec<WorldVector<double>, VERTEX>& getCoord() {
      return coord;    
    };

    /** \brief
     * Returns \ref boundary[i]
     */
    inline BoundaryType getBoundary(int i) const {
      return boundary[i]; 
    };

    /** \brief
     * Returns \ref elType
     */
    inline unsigned char getElType() const {
      return elType; 
    };

    /** \} */

    // ===== setting methods ======================================================

    /** \name setting methods
     * \{
     */

    /** \brief
     * Sets \ref index
     */
    inline void setIndex(int n) {
      index = n ; 
    };

    /** \brief
     * Sets \ref element if not yet set.
     */
    inline void setElement(Element* element_) {
      if (!element) {
	element = element_; 
      } else {
	if (element != element_) 
	  ERROR("Trying to change element in MacroElement\n");   
      }
    };

    /** \brief
     * Sets \ref elType
     */
    inline void setElType(unsigned char typ) {
      elType = typ; 
    };


    /** \brief
     * Sets \ref projection[i] = p.
     */
    inline void setProjection(int i, Projection *p) {
      projection[i] = p;
    };

    /** \brief
     * Sets the i-th Neighbour to n
     */
    inline void setNeighbour(int i, MacroElement *n) {
      neighbour[i] = n;
    };

    /** \brief
     * Sets the i-th opp vertex to c
     */
    inline void  setOppVertex(int i, char c) {
      oppVertex[i] = c;
    };

    /** \brief
     * Sets \ref boundary[i] to b
     */
    inline void setBoundary(int i, BoundaryType b) {
      boundary[i] = b; 
    };

    inline void setCoord(int i, const WorldVector<double> &c) {
      coord[i] = c;
    };

    /** \} */

    // ===== Serializable implementation =====

    void serialize(std::ostream &out);

    void deserialize(std::istream &in);

    inline void writeNeighboursTo(std::vector<int> *indices) {
      deserializedNeighbourIndices_ = indices;
    }

    int calcMemoryUsage();

  protected:
    /** \brief
     * Element of this MacroElement.
     */
    Element *element;

    /** \brief
     * Coordinates of the vertices.
     */
    FixVec<WorldVector<double>, VERTEX> coord;

    /** \brief
     * Boundary type of each boundary part (face/edge/vertex)
     */
    FixVec<BoundaryType, BOUNDARY> boundary;

    /** \brief
     * Boundary projection to curved boundaries or element projections.
     */
    FixVec<Projection*, PROJECTION> projection;

    /** \brief
     * Pointers to all neighbours of this MacroElement 
     */
    FixVec<MacroElement*, NEIGH> neighbour;

    /** \brief
     * opp vertices of this MacroElement
     */
    FixVec<char, NEIGH> oppVertex;

    /** \brief
     * index of this MacroElement
     */
    int index;  

    /** \brief
     * Element type of the MacroElement
     */
    unsigned char elType;  

    /** \brief
     *
     */
    std::vector<int> *deserializedNeighbourIndices_;

    // friend classes
    friend class MacroInfo;
    friend class MacroReader;
    friend class ElInfo1d;
    friend class ElInfo2d;
    friend class ElInfo3d;
  };

}


#endif  // AMDIS_MACROELEMENT_H 
